#include "liquidity_defs.h"
#include "realtype.h"
#include "tauchen.h"
#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <algorithm>
#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include "normaldist.h"

// #define DEBUG_IDENTITY_PROB

// preferences
__constant__ REAL_TYPE RRA;
__constant__ REAL_TYPE BETA;
__constant__ REAL_TYPE ONE_MINUS_RRA; // 1-RRA
__constant__ REAL_TYPE U_SCALING; // 1/(1-RRA)

// numerical parameters
__constant__ REAL_TYPE C_LB;
__constant__ REAL_TYPE GGQ_MLB;
__constant__ REAL_TYPE GGQ_MUB;
__constant__ REAL_TYPE GGQ_MMEAN;
__constant__ REAL_TYPE GGQ_MSTD;
__constant__ REAL_TYPE GGQ_MMASS;
__constant__ REAL_TYPE GGQ_MDEF;
__constant__ REAL_TYPE ERRTOL_BISECT;
__constant__ REAL_TYPE MAX_DEF_PROB;
__constant__ int MAXITERS_BISECT;
__constant__ int CONV_CHK_TFP_LB_IDX;
__constant__ int CONV_CHK_TFP_UB_IDX;
__constant__ int CONV_CHK_B_UB_IDX;

// liquidity discounts
__constant__ REAL_TYPE RH;
__constant__ REAL_TYPE RL;
__constant__ REAL_TYPE HOLDING_COST;

//// effective transitions

__constant__ REAL_TYPE PROB_HL;
__constant__ REAL_TYPE PROB_LH_ND;
__constant__ REAL_TYPE PROB_LH_D;
/*
Note:
1. H->L is Prob(get liquidity shock)
2. L->H is Prob(meet mkt maker)*(1-(mkt maker barg power))
*/

// debt market
__constant__ REAL_TYPE REENTERPROB;
__constant__ REAL_TYPE DEBT_Z;
__constant__ REAL_TYPE DEBT_M;

// grid dimensions
__constant__ int GRIDSIZE_Y;
__constant__ int GRIDSIZE_B;

// grids
__constant__ REAL_TYPE GRID_Y_ND[MAX_GRIDSIZE_Y];
__constant__ REAL_TYPE GRID_B[MAX_GRIDSIZE_B];

// interpolation
__constant__ int GRID_B_REENTER_IDX[MAX_GRIDSIZE_B];
__constant__ REAL_TYPE GRID_RECOVERY_FRAC[MAX_GRIDSIZE_B];

void trimstring(std::string &str, std::string trim_delimiters)
{
	str.erase(0, str.find_first_not_of(trim_delimiters));
	str.substr(0, str.find_last_not_of(trim_delimiters) + 1);
}

void delimStrToVec(REAL_TYPE *z, int &N, std::string s, char delim)
{
	// main body: REAL_TYPE *x;
	REAL_TYPE *x, *y;
	y = new REAL_TYPE[0];
	N = 0;
	std::stringstream ss(s);
	std::string item;

	while (std::getline(ss, item, delim))
	{
		// first copy x to a temp variable
		if (N>0)
		{
			delete[] y;
			y = new REAL_TYPE[N];
			for (int i = 0; i<N; i++)
			{
				y[i] = x[i];
			}
			delete[] x;
		}
		N += 1;
		x = new REAL_TYPE[N];
		for (int i = 0; i<N - 1; i++)
		{
			x[i] = y[i];
		}
		x[N - 1] = atof(item.c_str());
	}

	// copy to final
	// z = new REAL_TYPE[N];
	for (int i = 0; i<N; i++)
	{
		z[i] = x[i];
	}

	// deallocated memory to temp variables
	delete[] x;
	delete[] y;

}

void delimStrToVec(int *z, int &N, std::string s, char delim)
{
	// main body: REAL_TYPE *x;
	int *x, *y;
	y = new int[0];
	N = 0;
	std::stringstream ss(s);
	std::string item;

	while (std::getline(ss, item, delim))
	{
		// first copy x to a temp variable
		if (N>0)
		{
			delete[] y;
			y = new int[N];
			for (int i = 0; i<N; i++)
			{
				y[i] = x[i];
			}
			delete[] x;
		}
		N += 1;
		x = new int[N];
		for (int i = 0; i<N - 1; i++)
		{
			x[i] = y[i];
		}
		x[N - 1] = atoi(item.c_str());
	}

	// copy to final
	// z = new int[N];
	for (int i = 0; i<N; i++)
	{
		z[i] = x[i];
	}

	// deallocated memory to temp variables
	delete[] x;
	delete[] y;

}

REAL_TYPE CalcAutarkyOutput(REAL_TYPE yND, REAL_TYPE bdef, REAL_TYPE d_y, REAL_TYPE d_yy, REAL_TYPE d_b, REAL_TYPE d_bb, REAL_TYPE d_yb)
{
	REAL_TYPE yD = yND - min(yND, max(0.0, d_y*yND + d_yy*yND*yND + d_b*bdef + d_bb*bdef*bdef + d_yb*yND*bdef));
	return yD;
}

void parms_bsl_mod::clean()
{
	if (grid_y_nd)
	{
		delete[] grid_y_nd;
		grid_y_nd = NULL;
	}
	if (grid_y_d)
	{
		delete[] grid_y_d;
		grid_y_d = NULL;
	}
	if (grid_b)
	{
		delete[] grid_b;
		grid_b = NULL;
	}
	if (prob_y)
	{
		delete[] prob_y;
		prob_y = NULL;
	}
	if (grid_uD_yb)
	{
		delete[] grid_uD_yb;
		grid_uD_yb = NULL;
	}
	if (UpdateRegimeIter)
	{
		delete[] UpdateRegimeIter;
		UpdateRegimeIter = NULL;
	}
	if (UpdateWeightOldV)
	{
		delete[] UpdateWeightOldV;
		UpdateWeightOldV = NULL;
	}
	if (UpdateWeightOldQ)
	{
		delete[] UpdateWeightOldQ;
		UpdateWeightOldQ = NULL;
	}
	if (UpdateWeightOldDefProb)
	{
		delete[] UpdateWeightOldDefProb;
		UpdateWeightOldDefProb = NULL;
	}
}

int parms_bsl_mod::checkparms()
{
	bool bParmsAllOK = true;
	if (rra <= 0)
	{
		std::cout << "warning: rra<=0." << std::endl;
		bParmsAllOK = false;
	}
	if (beta<0 || beta>=1)
	{
		std::cout << "warning: beta<0 || beta>=1." << std::endl;
		bParmsAllOK = false;
	}
	if (clb <= 0)
	{
		std::cout << "warning: clb<=0." << std::endl;
		bParmsAllOK = false;
	}
	if (ggq_mlb >= ggq_mub)
	{
		std::cout << "warning: ggq_mlb >= ggq_mub." << std::endl;
		bParmsAllOK = false;
	}
	if (ggq_mDval < ggq_mlb || ggq_mDval > ggq_mub)
	{
		std::cout << "warning: ggq_mDval < ggq_mlb || ggq_mDval > ggq_mub." << std::endl;
		bParmsAllOK = false;
	}
	if (ggq_mmean < ggq_mlb || ggq_mmean > ggq_mub)
	{
		std::cout << "warning: ggq_mmean < ggq_mlb || ggq_mmean > ggq_mub." << std::endl;
		bParmsAllOK = false;
	}
	if (ggq_mstd <= 0 )
	{
		std::cout << "warning: ggq_mstd <= 0." << std::endl;
		bParmsAllOK = false;
	}
	if (errtol_bisect <= 0)
	{
		std::cout << "warning: errtol_bisect <= 0." << std::endl;
		bParmsAllOK = false;
	}
	if (maxiters_bisect <= 0)
	{
		std::cout << "warning: maxiters_bisect <= 0." << std::endl;
		bParmsAllOK = false;
	}
	if (rH > rL)
	{
		std::cout << "warning: rH > rL." << std::endl;
		bParmsAllOK = false;
	}
	if (rH < 0 || rL < 0)
	{
		std::cout << "warning: rH < 0 || rL < 0." << std::endl;
		bParmsAllOK = false;
	}
	if (prob_liqshock<0 || prob_liqshock>1 ||
		prob_meet_mkt_maker_ND<0 || prob_meet_mkt_maker_ND>1 ||
		prob_meet_mkt_maker_D < 0 || prob_meet_mkt_maker_D>1 ||
		bargain_power_mkt_maker_ND<0 || bargain_power_mkt_maker_ND>1 ||
		bargain_power_mkt_maker_D<0 || bargain_power_mkt_maker_D>1)
	{
		std::cout << "warning: probabilities from liquidity transitions not valid." << std::endl;
		bParmsAllOK = false;
	}

	if (reenterprob < 0 || reenterprob>1)
	{
		std::cout << "warning: reenterprob not valid." << std::endl;
		bParmsAllOK = false;
	}

	if (debt_z < 0 || debt_m<0 || debt_m>1 ||  
		debt_writedown_reentry<0 || debt_writedown_reentry>1 || max_def_prob<0 || max_def_prob>1 || recovery_bmax<0)
	{
		std::cout << "warning: invalid debt parameters." << std::endl;
		bParmsAllOK = false;
	}

	if (gridsize_tfp<2 || gridsize_b<2)
	{
		std::cout << "warning: too few grid points." << std::endl;
		bParmsAllOK = false;
	}

	if (gridsize_tfp > MAX_GRIDSIZE_Y || gridsize_b>MAX_GRIDSIZE_B)
	{
		std::cout << "warning: grid size exceeds maximum allowed (256 for y, 512 for b)" << std::endl;
		bParmsAllOK = false;
	}

	if (bmax<=0 || 0>= conv_chk_b_ub || conv_chk_b_ub>bmax)
	{
		std::cout << "warning: bmax<=0." << std::endl;
		bParmsAllOK = false;
	}

	if (rho_tfp<=0 || rho_tfp>=1 || sigma_tfp<=0 || tfp_lb>=0 || tfp_ub<=0 || conv_chk_tfp_lb>=conv_chk_tfp_ub)
	{
		std::cout << "warning: invalid tfp dynamics." << std::endl;
		bParmsAllOK = false;
	}

	bool bValidUpdateWeights = true;
	for (int i = 0; i < NumUpdateRegimes; i++)
	{
		if (UpdateWeightOldV[i]<0 || UpdateWeightOldV[i]>1 ||
			UpdateWeightOldQ[i]<0 || UpdateWeightOldQ[i]>1 ||
			UpdateWeightOldDefProb[i]<0 || UpdateWeightOldDefProb[i]>1)
		{
			bValidUpdateWeights = false;
			break;
		}
	}
	if (!bValidUpdateWeights)
	{
		std::cout << "warning: invalid update weights detected." << std::endl;
		bParmsAllOK = false;
	}
	
	if (bParmsAllOK)
	{
		std::cout << "Parameters satisfy checks." << std::endl;
		return EXIT_SUCCESS;
	}
	else
	{
		std::cout << "Parameters do not satisfy checks." << std::endl;
		return EXIT_FAILURE;
	}

}

void parms_bsl_mod::initDeviceConstants()
{
	REAL_TYPE tempf;
	// preferences
	cudaMemcpyToSymbol(RRA, &rra, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(BETA, &beta, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	tempf = 1 - rra;
	cudaMemcpyToSymbol(ONE_MINUS_RRA, &tempf, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	tempf = 1/(1 - rra);
	cudaMemcpyToSymbol(U_SCALING, &tempf, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	// numerical parameters
	cudaMemcpyToSymbol(C_LB, &clb, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(GGQ_MLB, &ggq_mlb, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(GGQ_MUB, &ggq_mub, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(GGQ_MMEAN, &ggq_mmean, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(GGQ_MSTD, &ggq_mstd, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(GGQ_MMASS, &ggq_mmass, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(GGQ_MDEF, &ggq_mDval, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(ERRTOL_BISECT, &errtol_bisect, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(MAXITERS_BISECT, &maxiters_bisect, 1 * sizeof(int), 0, cudaMemcpyHostToDevice);
	// liquidity discounts
	cudaMemcpyToSymbol(RH, &rH, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(RL, &rL, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(HOLDING_COST, &holding_cost, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	// effective liquidity transitions
	cudaMemcpyToSymbol(PROB_HL, &prob_liqshock, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	tempf = prob_meet_mkt_maker_ND*(1- bargain_power_mkt_maker_ND);
	cudaMemcpyToSymbol(PROB_LH_ND, &tempf, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	tempf = prob_meet_mkt_maker_D*(1 - bargain_power_mkt_maker_D);
	cudaMemcpyToSymbol(PROB_LH_D, &tempf, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	// debt market
	cudaMemcpyToSymbol(REENTERPROB, &reenterprob, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(DEBT_Z, &debt_z, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(DEBT_M, &debt_m, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(MAX_DEF_PROB, &max_def_prob, 1 * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	// grid dimensions
	cudaMemcpyToSymbol(GRIDSIZE_Y, &gridsize_tfp, 1 * sizeof(int), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(GRIDSIZE_B, &gridsize_b, 1 * sizeof(int), 0, cudaMemcpyHostToDevice);

	cudaMemcpyToSymbol(GRID_Y_ND, grid_y_nd, gridsize_tfp * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);

	cudaMemcpyToSymbol(GRID_B, grid_b, gridsize_b * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);

	// nearest point interpolation for reentry debt
	int *grid_b_reenter_idx = new int[gridsize_b];
	int tempidx;
	REAL_TYPE tempdiff;
	REAL_TYPE *grid_recovery_frac = new REAL_TYPE[gridsize_b];
	
	for (int idx_b = 0; idx_b<gridsize_b; idx_b++)
	{
		tempidx = 0;
		tempf = (idx_b>0)?grid_b[idx_b]*min(1 - debt_writedown_reentry,recovery_bmax/grid_b[idx_b]):grid_b[idx_b]*(1 - debt_writedown_reentry);
		tempdiff = abs(grid_b[0] - tempf);
		for (int j = 1; j<gridsize_b; j++)
		{
			if (abs(grid_b[j] - tempf) < tempdiff)
			{
				tempdiff = abs(grid_b[j] - tempf);
				tempidx = j;
			}
		}
		grid_b_reenter_idx[idx_b] = tempidx;
		grid_recovery_frac[idx_b] = (idx_b>0)?min(1 - debt_writedown_reentry,recovery_bmax/grid_b[idx_b]):(1 - debt_writedown_reentry);
	}

	cudaMemcpyToSymbol(GRID_B_REENTER_IDX, grid_b_reenter_idx, gridsize_b * sizeof(int), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(GRID_RECOVERY_FRAC, grid_recovery_frac, gridsize_b * sizeof(REAL_TYPE), 0, cudaMemcpyHostToDevice);
	delete[] grid_b_reenter_idx;
	delete[] grid_recovery_frac;

	cudaDeviceSynchronize();

}

int parms_bsl_mod::readparms(char *filename)
{
	char delimiter = ':';
	std::string tempstr;

	if (GetFieldFromFile(tempstr, filename, "rra", delimiter))
	{
		// rra = std::stof(tempstr);
		rra = atof(tempstr.c_str());
	}
	else
	{
		printf("rra not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "beta", delimiter))
	{
		// beta = std::stof(tempstr);
		beta = atof(tempstr.c_str());
	}
	else
	{
		printf("beta not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "clb", delimiter))
	{
		// clb = std::stof(tempstr);
		clb = atof(tempstr.c_str());
	}
	else
	{
		printf("clb not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "ggq_mDval", delimiter))
	{
		// ggq_mDval = std::stof(tempstr);
		ggq_mDval = atof(tempstr.c_str());
	}
	else
	{
		printf("ggq_mDval not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "ggq_mlb", delimiter))
	{
		// ggq_mlb = std::stof(tempstr);
		ggq_mlb = atof(tempstr.c_str());
	}
	else
	{
		printf("ggq_mlb not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "ggq_mub", delimiter))
	{
		// ggq_mub = std::stof(tempstr);
		ggq_mub = atof(tempstr.c_str());
	}
	else
	{
		printf("ggq_mub not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "ggq_mmean", delimiter))
	{
		// ggq_mmean = std::stof(tempstr);
		ggq_mmean = atof(tempstr.c_str());
	}
	else
	{
		printf("ggq_mmean not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "ggq_mstd", delimiter))
	{
		// ggq_mstd = std::stof(tempstr);
		ggq_mstd = atof(tempstr.c_str());
	}
	else
	{
		printf("ggq_mstd not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	ggq_mmass = normcdf(ggq_mub - ggq_mmean,ggq_mstd) - normcdf(ggq_mlb - ggq_mmean, ggq_mstd);

	if (GetFieldFromFile(tempstr, filename, "max_def_prob", delimiter))
	{
		max_def_prob = atof(tempstr.c_str());
	}
	else
	{
		printf("max_def_prob not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "d_y", delimiter))
	{
		// d_y = std::stof(tempstr);
		d_y = atof(tempstr.c_str());
	}
	else
	{
		printf("d_y not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "d_yy", delimiter))
	{
		// d_yy = std::stof(tempstr);
		d_yy = atof(tempstr.c_str());
	}
	else
	{
		printf("d_yy not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "d_b", delimiter))
	{
		// d_b = std::stof(tempstr);
		d_b = atof(tempstr.c_str());
	}
	else
	{
		printf("d_b not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "d_bb", delimiter))
	{
		// d_bb = std::stof(tempstr);
		d_bb = atof(tempstr.c_str());
	}
	else
	{
		printf("d_bb not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "d_yb", delimiter))
	{
		// d_yb = std::stof(tempstr);
		d_yb = atof(tempstr.c_str());
	}
	else
	{
		printf("d_yb not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "errtol_bisect", delimiter))
	{
		errtol_bisect = atof(tempstr.c_str());
	}
	else
	{
		printf("errtol_bisect not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "maxiters_bisect", delimiter))
	{
		maxiters_bisect = atoi(tempstr.c_str());
	}
	else
	{
		printf("maxiters_bisect not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "rH", delimiter))
	{
		// rH_ND = std::stof(tempstr);
		rH = atof(tempstr.c_str());
	}
	else
	{
		printf("rH not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "rL", delimiter))
	{
		rL = atof(tempstr.c_str());
	}
	else
	{
		printf("rL not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "holding_cost", delimiter))
	{
		holding_cost = atof(tempstr.c_str());
	}
	else
	{
		printf("holding_cost not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "prob_liqshock", delimiter))
	{
		prob_liqshock = atof(tempstr.c_str());
	}
	else
	{
		printf("prob_liqshock not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "prob_meet_mkt_maker_ND", delimiter))
	{
		prob_meet_mkt_maker_ND = atof(tempstr.c_str());
	}
	else
	{
		printf("prob_meet_mkt_maker_ND not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "prob_meet_mkt_maker_D", delimiter))
	{
		prob_meet_mkt_maker_D = atof(tempstr.c_str());
	}
	else
	{
		printf("prob_meet_mkt_maker_D not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "bargain_power_mkt_maker_ND", delimiter))
	{
		// bargain_power_mkt_maker_ND = std::stof(tempstr);
		bargain_power_mkt_maker_ND = atof(tempstr.c_str());
	}
	else
	{
		printf("bargain_power_mkt_maker_ND not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "bargain_power_mkt_maker_D", delimiter))
	{
		// bargain_power_mkt_maker_D = std::stof(tempstr);
		bargain_power_mkt_maker_D = atof(tempstr.c_str());
	}
	else
	{
		printf("bargain_power_mkt_maker_D not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "reenterprob", delimiter))
	{
		//reenterprob = std::stof(tempstr);
		reenterprob = atof(tempstr.c_str());
	}
	else
	{
		printf("reenterprob not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "debt_z", delimiter))
	{
		// debt_z = std::stof(tempstr);
		debt_z = atof(tempstr.c_str());
	}
	else
	{
		printf("debt_z not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "debt_m", delimiter))
	{
		// debt_m = std::stof(tempstr);
		debt_m = atof(tempstr.c_str());
	}
	else
	{
		printf("debt_m not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "debt_writedown_reentry", delimiter))
	{
		// debt_writedown_reentry = std::stof(tempstr);
		debt_writedown_reentry = atof(tempstr.c_str());
	}
	else
	{
		printf("debt_writedown_reentry not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "recovery_bmax", delimiter))
	{
		// debt_writedown_reentry = std::stof(tempstr);
		recovery_bmax = atof(tempstr.c_str());
	}
	else
	{
		printf("recovery_bmax not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "gridsize_tfp", delimiter))
	{
		// gridsize_tfp = std::stoi(tempstr);
		gridsize_tfp = atoi(tempstr.c_str());
	}
	else
	{
		printf("gridsize_tfp not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "gridsize_b", delimiter))
	{
		//gridsize_b = std::stoi(tempstr);
		gridsize_b = atoi(tempstr.c_str());
	}
	else
	{
		printf("gridsize_b not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "bmax", delimiter))
	{
		// bmax = std::stof(tempstr);
		bmax = atof(tempstr.c_str());
	}
	else
	{
		printf("bmax not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "sigma_tfp", delimiter))
	{
		// sigma_tfp = std::stof(tempstr);
		sigma_tfp = atof(tempstr.c_str());
	}
	else
	{
		printf("sigma_tfp not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "rho_tfp", delimiter))
	{
		// rho_tfp = std::stof(tempstr);
		rho_tfp = atof(tempstr.c_str());
	}
	else
	{
		printf("rho_tfp not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "tfp_lb", delimiter))
	{
		tfp_lb = atof(tempstr.c_str());
	}
	else
	{
		printf("tfp_lb not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "tfp_ub", delimiter))
	{
		tfp_ub = atof(tempstr.c_str());
	}
	else
	{
		printf("tfp_ub not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "conv_chk_tfp_lb", delimiter))
	{
		conv_chk_tfp_lb = atof(tempstr.c_str());
	}
	else
	{
		printf("conv_chk_tfp_lb not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "conv_chk_tfp_ub", delimiter))
	{
		conv_chk_tfp_ub = atof(tempstr.c_str());
	}
	else
	{
		printf("conv_chk_tfp_ub not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "conv_chk_b_ub", delimiter))
	{
		conv_chk_b_ub = atof(tempstr.c_str());
	}
	else
	{
		printf("conv_chk_b_ub not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "maxiters", delimiter))
	{
		// maxiters = std::stoi(tempstr);
		maxiters = atoi(tempstr.c_str());
	}
	else
	{
		printf("maxiters not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "errtolV", delimiter))
	{
		// errtolV = std::stof(tempstr);
		errtolV = atof(tempstr.c_str());
	}
	else
	{
		printf("errtolV not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "errtolQ", delimiter))
	{
		// errtolQ = std::stof(tempstr);
		errtolQ = atof(tempstr.c_str());
	}
	else
	{
		printf("errtolQ not found in %s\n", filename);
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "NumUpdateRegimes", delimiter))
	{
		// NumUpdateRegimes = std::stoi(tempstr);
		NumUpdateRegimes = atoi(tempstr.c_str());
	}
	else
	{
		printf("NumUpdateRegimes not found");
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "UpdateRegimeIter", delimiter))
	{
		int N;
		UpdateRegimeIter = new int[NumUpdateRegimes - 1];
		delimStrToVec(UpdateRegimeIter, N, tempstr, '/');
		if (N != NumUpdateRegimes - 1)
		{
			printf("UpdateRegimeIter has wrong number of inputs");
			exit(EXIT_FAILURE);
		}
	}
	else
	{
		printf("UpdateRegimeIter not found");
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "UpdateWeightOldV", delimiter))
	{
		int N;
		UpdateWeightOldV = new REAL_TYPE[NumUpdateRegimes];
		delimStrToVec(UpdateWeightOldV, N, tempstr, '/');
		if (N != NumUpdateRegimes)
		{
			printf("UpdateWeightOldV has wrong number of inputs");
			exit(EXIT_FAILURE);
		}
	}
	else
	{
		printf("UpdateWeightOldV not found");
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "UpdateWeightOldQ", delimiter))
	{
		int N;
		UpdateWeightOldQ = new REAL_TYPE[NumUpdateRegimes];
		delimStrToVec(UpdateWeightOldQ, N, tempstr, '/');
		if (N != NumUpdateRegimes)
		{
			printf("UpdateWeightOldQ has wrong number of inputs");
			exit(EXIT_FAILURE);
		}
	}
	else
	{
		printf("UpdateWeightOldQ not found");
		exit(EXIT_FAILURE);
	}

	if (GetFieldFromFile(tempstr, filename, "UpdateWeightOldDefProb", delimiter))
	{
		int N;
		UpdateWeightOldDefProb = new REAL_TYPE[NumUpdateRegimes];
		delimStrToVec(UpdateWeightOldDefProb, N, tempstr, '/');
		if (N != NumUpdateRegimes)
		{
			printf("UpdateWeightOldDefProb has wrong number of inputs");
			exit(EXIT_FAILURE);
		}
	}
	else
	{
		printf("UpdateWeightOldDefProb not found");
		exit(EXIT_FAILURE);
	}

	// set up the grid for b
	grid_b = new REAL_TYPE[gridsize_b];
	REAL_TYPE db = bmax / (REAL_TYPE)(gridsize_b - 1);
	for (int i = 0; i < gridsize_b; i++)
	{
		grid_b[i] = db*(REAL_TYPE)i;
	}

	// use tauchen
	REAL_TYPE *z, *prob_z;
	z = new REAL_TYPE[gridsize_tfp];
	prob_z = new REAL_TYPE[gridsize_tfp*gridsize_tfp];
	
	if (tauchen(z, prob_z, rho_tfp, sigma_tfp, gridsize_tfp, tfp_lb, tfp_ub) == EXIT_SUCCESS)
	{
		grid_y_nd = new REAL_TYPE[gridsize_tfp];
		grid_y_d = new REAL_TYPE[gridsize_tfp*gridsize_b];
		grid_uD_yb = new REAL_TYPE[gridsize_tfp*gridsize_b];
		prob_y = new REAL_TYPE[gridsize_tfp*gridsize_tfp];
		for (int i = 0; i < gridsize_tfp; i++)
		{
			grid_y_nd[i] = exp(z[i]);
			for (int j = 0; j < gridsize_tfp; j++)
			{
				prob_y[i*gridsize_tfp + j] = prob_z[i*gridsize_tfp + j];
			}
			for (int k = 0; k < gridsize_b; k++)
			{
				grid_y_d[i*gridsize_b + k] = CalcAutarkyOutput(grid_y_nd[i], grid_b[k], d_y, d_yy, d_b, d_bb, d_yb);
				grid_uD_yb[i*gridsize_b + k] = POWERFUN(max(clb, grid_y_d[i*gridsize_b + k]),1-rra)/(1-rra);
			}
		}
	}
	else
	{
		std::cout << "Error running rouwenhorst." << std::endl;
		exit(EXIT_FAILURE);
	}

	#ifdef DEBUG_IDENTITY_PROB
		for (int i = 0; i < gridsize_tfp; i++)
		{
			for (int j = 0; j < gridsize_tfp; j++)
			{
				if (i==j)
				{
					prob_y[i*gridsize_tfp + j] = 1.0;
				}
				else
				{
					prob_y[i*gridsize_tfp + j] = 0.0;
				}
			}
		}
	#endif

	// compute CONV_CHK_TFP_LB_IDX and CONV_CHK_TFP_UB_IDX;
	int conv_chk_tfp_lb_idx, conv_chk_tfp_ub_idx, conv_chk_b_ub_idx;
	for (int i = 0; i < gridsize_tfp; i++)
	{
		if (z[i]>=conv_chk_tfp_lb)
		{
			conv_chk_tfp_lb_idx = i;
			break;
		}
	}
	for (int i = 0; i < gridsize_tfp; i++)
	{
		if (z[i]<=conv_chk_tfp_ub)
		{
			conv_chk_tfp_ub_idx = i;
		}
		else
		{
			break;
		}
	}
	for (int i = 0; i < gridsize_b; i++)
	{
		if (grid_b[i] <= conv_chk_b_ub)
		{
			conv_chk_b_ub_idx = i;
		}
		else
		{
			break;
		}
	}

	delete[] z;
	delete[] prob_z;

	cudaMemcpyToSymbol(CONV_CHK_TFP_LB_IDX, &conv_chk_tfp_lb_idx, 1 * sizeof(int), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(CONV_CHK_TFP_UB_IDX, &conv_chk_tfp_ub_idx, 1 * sizeof(int), 0, cudaMemcpyHostToDevice);
	cudaMemcpyToSymbol(CONV_CHK_B_UB_IDX, &conv_chk_b_ub_idx, 1 * sizeof(int), 0, cudaMemcpyHostToDevice);

	return EXIT_SUCCESS;

}

bool GetFieldFromFile(std::string &out_str, char *filename, std::string fieldname_in, char delimiter)
/*
format: fieldname delimiter value
*/
{

	std::ifstream ifs;
	std::size_t foundpos, foundpos2;

	ifs.open(filename);

	std::string trim_delimiters = " \f\n\r\t\v";

	bool bFoundStr = false;

	if (ifs.is_open())
	{

		std::string fieldname = fieldname_in;
		std::transform(fieldname.begin(), fieldname.end(), fieldname.begin(), ::tolower);

		std::string str, str1, str2;

		bool bContinue = true;

		while (bContinue)
		{

			std::getline(ifs, str);
			foundpos = str.find(delimiter);

			if (foundpos != std::string::npos)
			{
				// check that there's a single delimiter
				foundpos2 = str.find(delimiter, foundpos + 1);
				if (foundpos2 == std::string::npos)
				{
					// only a single delimiter found
					str1 = str.substr(0, foundpos);
					trimstring(str1);
					std::transform(str1.begin(), str1.end(), str1.begin(), ::tolower);
					if (str1.compare(fieldname) == 0)
					{
						str2 = str.substr(foundpos + 1);
						trimstring(str2);
						out_str = str2;
						bContinue = false;
						bFoundStr = true;
					}
				}
			}

			if (ifs.eof())
			{
				bContinue = false;
			}
		}

		ifs.close();

	}

	return bFoundStr;

}

void parms_bsl_mod::displayparms()
{
	std::cout << "Parameters: " << std::endl;
	std::cout << "rra: " << rra << std::endl;
	std::cout << "beta: " << beta << std::endl;
	std::cout << "clb: " << clb << std::endl;
	std::cout << "ggq_mDval: " << ggq_mDval << std::endl;
	std::cout << "ggq_mlb: " << ggq_mlb << std::endl;
	std::cout << "ggq_mub: " << ggq_mub << std::endl;
	std::cout << "ggq_mmean: " << ggq_mmean << std::endl;
	std::cout << "ggq_mstd: " << ggq_mstd << std::endl;
	std::cout << "ggq_mmass: " << ggq_mmass << std::endl;
	std::cout << "errtol_bisect: " << errtol_bisect << std::endl;
	std::cout << "maxiters_bisect: " << maxiters_bisect << std::endl;
	std::cout << "rH: " << rH << std::endl;
	std::cout << "rL: " << rL << std::endl;
	std::cout << "holding_cost: " << holding_cost << std::endl;
	std::cout << "prob_liqshock: " << prob_liqshock << std::endl;
	std::cout << "prob_meet_mkt_maker_ND: " << prob_meet_mkt_maker_ND << std::endl;
	std::cout << "prob_meet_mkt_maker_D: " << prob_meet_mkt_maker_D << std::endl;
	std::cout << "bargain_power_mkt_maker_ND: " << bargain_power_mkt_maker_ND << std::endl;
	std::cout << "bargain_power_mkt_maker_D: " << bargain_power_mkt_maker_D << std::endl;
	std::cout << "reenterprob: " << reenterprob << std::endl;
	std::cout << "debt_z: " << debt_z << std::endl;
	std::cout << "debt_m: " << debt_m << std::endl;
	std::cout << "debt_writedown_reentry: " << debt_writedown_reentry << std::endl;
	std::cout << "max_def_prob: " << max_def_prob << std::endl;
	std::cout << "gridsize_tfp: " << gridsize_tfp << std::endl;
	std::cout << "gridsize_b: " << gridsize_b << std::endl;
	std::cout << "bmax: " << bmax << std::endl;
	std::cout << "tfp_lb: " << tfp_lb << std::endl;
	std::cout << "tfp_ub: " << tfp_ub << std::endl;
	std::cout << "conv_chk_tfp_lb: " << conv_chk_tfp_lb << std::endl;
	std::cout << "conv_chk_tfp_ub: " << conv_chk_tfp_ub << std::endl;
	std::cout << "conv_chk_b_ub: " << conv_chk_b_ub << std::endl;
	std::cout << "rho_tfp: " << rho_tfp << std::endl;
	std::cout << "sigma_tfp: " << sigma_tfp << std::endl;
	std::cout << "d_y: " << d_y << std::endl;
	std::cout << "d_yy: " << d_yy << std::endl;
	std::cout << "d_b: " << d_b << std::endl;
	std::cout << "d_bb: " << d_bb << std::endl;
	std::cout << "d_yb: " << d_yb << std::endl;
	std::cout << "maxiters: " << maxiters << std::endl;
	std::cout << "errtolV: " << errtolV << std::endl;
	std::cout << "errtolQ: " << errtolQ << std::endl;
	std::cout << "NumUpdateRegimes: " << NumUpdateRegimes << std::endl;
	for (int i = 0; i < NumUpdateRegimes - 1; i++)
	{
		std::cout << "UpdateRegimeIter[" << i << "]: " << UpdateRegimeIter[i] << std::endl;
	}
	for (int i = 0; i < NumUpdateRegimes; i++ )
	{
		std::cout << "UpdateWeightOldV[" << i << "]: " << UpdateWeightOldV[i] << std::endl;
		std::cout << "UpdateWeightOldQ[" << i << "]: " << UpdateWeightOldQ[i] << std::endl;
		std::cout << "UpdateWeightOldDefProb[" << i << "]: " << UpdateWeightOldDefProb[i] << std::endl;
	}
}

__global__ void printDeviceConstants()
{
	int i = blockIdx.x*blockDim.x + threadIdx.x;
	if (i == 0)
	{
		printf("Printing device constants.\n");
		printf("RRA = %g.\n", RRA);
		printf("BETA = %g.\n", BETA);
		printf("ONE_MINUS_RRA = %g.\n", ONE_MINUS_RRA);
		printf("U_SCALING = %g.\n", U_SCALING);
		printf("C_LB = %g.\n", C_LB);
		printf("GGQ_MLB = %g.\n", GGQ_MLB);
		printf("GGQ_MUB = %g.\n", GGQ_MUB);
		printf("GGQ_MMEAN = %g.\n", GGQ_MMEAN);
		printf("GGQ_MSTD = %g.\n", GGQ_MSTD);
		printf("GGQ_MMASS = %g.\n", GGQ_MMASS);
		printf("GGQ_MDEF = %g.\n", GGQ_MDEF);
		printf("ERRTOL_BISECT = %g.\n", ERRTOL_BISECT);
		printf("MAXITERS_BISECT = %d.\n", MAXITERS_BISECT);
		printf("RH = %g.\n", RH);
		printf("RL = %g.\n", RL);
		printf("HOLDING_COST = %g.\n", HOLDING_COST);
		printf("PROB_HL = %g.\n", PROB_HL);
		printf("PROB_LH_ND = %g.\n", PROB_LH_ND);
		printf("PROB_LH_D = %g.\n", PROB_LH_D);
		printf("REENTERPROB = %g.\n", REENTERPROB);
		printf("DEBT_Z = %g.\n", DEBT_Z);
		printf("DEBT_M = %g.\n", DEBT_M);
		printf("MAX_DEF_PROB = %g.\n", MAX_DEF_PROB);
		printf("GRIDSIZE_Y = %d.\n", GRIDSIZE_Y);
		printf("GRIDSIZE_B = %d.\n", GRIDSIZE_B);
		// for (i = 0; i < GRIDSIZE_Y; i++)
		// {
		// 	printf("GRID_Y_ND[%d]=%g.\n",i, GRID_Y_ND[i]);
		// }
		// for (i = 0; i < GRIDSIZE_B; i++)
		// {
		// 	printf("GRID_B[%d]=%g.\n",i, GRID_B[i]);
		// }
		// for (i = 0; i < GRIDSIZE_B; i++)
		// {
		// 	printf("i=%d. writedown interp: pt1=%d, wt1=%g.\n", i, 
		// 		INTERP1_WRITEDOWN_REENTER_PT1[i], INTERP1_WRITEDOWN_REENTER_WT1[i]);
		// }
	}
}