#include "realtype.h"
#include "liquidity_vfi.h"
#include "liquidity_defs.h" // includes device constant definitions
#include "ggq.h"
#include "normaldist.h"
#include <cmath>
#include<thrust/device_vector.h>
#include<thrust/device_ptr.h>
#include<thrust/extrema.h>
#include <iostream>
#include <cstdlib>
#include <cstdio>
#include "liquidity_io.h"
#include <math_constants.h>
#include <cstring>

#define DEBUG_PRINTF

__global__ void printDeviceElement(REAL_TYPE *x, int idx)
{
	int i = blockIdx.x*blockDim.x + threadIdx.x;
	if (i == 0)
	{
		printf("printDeviceElement:: x[%d]=%g.\n", idx, x[idx]);
	}
}

/*
FORMATS:
1. VD(y,b), VND(y,b), q(y,b'), CVD(y,b), CVND(y,b')
*/


__global__ void vfi_iterate(REAL_TYPE *d_VD, REAL_TYPE *d_VNDupdate, REAL_TYPE *d_qH_ND_update, REAL_TYPE *d_qL_ND_update, 
	REAL_TYPE *d_defprob_update, REAL_TYPE *d_defthresh,
	REAL_TYPE *d_CVD, REAL_TYPE *d_qH_ND, REAL_TYPE *d_qL_ND, 
	REAL_TYPE *d_qH_D, REAL_TYPE *d_qL_D, REAL_TYPE *d_CVND, REAL_TYPE *d_uD_yb, REAL_TYPE *d_defprob)
// we don't store the debt policy here
{
	int i = blockIdx.x*blockDim.x + threadIdx.x;
	volatile int idx_b = i % GRIDSIZE_B;
	volatile int idx_y = (i - idx_b) / GRIDSIZE_B;
	if (idx_y < GRIDSIZE_Y && idx_b < GRIDSIZE_B)
	{
		// update D problem
		REAL_TYPE VD = d_uD_yb[idx_y*GRIDSIZE_B + idx_b] + BETA*d_CVD[idx_y*GRIDSIZE_B + idx_b];

		REAL_TYPE CENDupdate, qHupdate, qLupdate, defprobupdate, defthresh;

		// partition and apply ggq algorithm on each sub-region.
		ggq_topdown(CENDupdate, qHupdate, qLupdate, defprobupdate, defthresh,
				idx_y, idx_b, VD,
				d_qH_ND, d_qL_ND, d_qH_D, d_qL_D, d_CVND, d_defprob);

		d_VNDupdate[idx_y*GRIDSIZE_B + idx_b] = CENDupdate;
		d_qH_ND_update[idx_y*GRIDSIZE_B + idx_b] = qHupdate;
		d_qL_ND_update[idx_y*GRIDSIZE_B + idx_b] = qLupdate;
		d_defprob_update[idx_y*GRIDSIZE_B + idx_b] = defprobupdate;
		d_VD[idx_y*GRIDSIZE_B + idx_b] = VD;
		d_defthresh[idx_y*GRIDSIZE_B + idx_b] = defthresh;

	}
}

__global__ void vfi_interpolate(REAL_TYPE *ceCi, REAL_TYPE *qH_NDi, REAL_TYPE *qL_NDi,
	REAL_TYPE *ceC, REAL_TYPE *qH_ND, REAL_TYPE *qL_ND)
/*
interpolation based on writedown_rate:
ceD_haircut, qH_D, qL_D
interpolation based on writedown_rate and writedown_reenter:
ceC_haircut, qH_ND, qL_ND
*/
{
	int i = blockIdx.x*blockDim.x + threadIdx.x;
	int idx_b = i % GRIDSIZE_B;
	int idx_y = (i - idx_b) / GRIDSIZE_B;
	if (i < GRIDSIZE_Y*GRIDSIZE_B)
	{
		ceCi[i] = ceC[idx_y*GRIDSIZE_B + GRID_B_REENTER_IDX[idx_b]];
		qH_NDi[i] = qH_ND[idx_y*GRIDSIZE_B + GRID_B_REENTER_IDX[idx_b]];
		qL_NDi[i] = qL_ND[idx_y*GRIDSIZE_B + GRID_B_REENTER_IDX[idx_b]];
	}
}

__global__ void vfi_update1(REAL_TYPE *d_prob_y, REAL_TYPE *d_EVD, REAL_TYPE *d_ceC, 
	REAL_TYPE *d_qH_ND, REAL_TYPE *d_qL_ND, REAL_TYPE *d_qH_D1, REAL_TYPE *d_qL_D1, REAL_TYPE *d_Edefprob,
	REAL_TYPE *d_VD, REAL_TYPE *d_VNDupdate, REAL_TYPE *d_qH_NDupdate, REAL_TYPE *d_qL_NDupdate, REAL_TYPE *d_defprob_update, 
	REAL_TYPE *d_qH_D0, REAL_TYPE *d_qL_D0)
	// note: no interpolation occurs here
{
	int i = blockIdx.x*blockDim.x + threadIdx.x;
	volatile int idx_b = i % GRIDSIZE_B;
	volatile int idx_y = (i - idx_b) / GRIDSIZE_B;
	if (idx_y < GRIDSIZE_Y && idx_b < GRIDSIZE_B)
	{
		REAL_TYPE tempf;

		// EVD
		tempf = 0;
		for (i = 0; i < GRIDSIZE_Y; i++)
		{
			tempf += d_prob_y[idx_y*GRIDSIZE_Y + i] * d_VD[i*GRIDSIZE_B + idx_b];
		}
		d_EVD[idx_y*GRIDSIZE_B + idx_b] = tempf;

		// ceC
		tempf = 0;
		for (i = 0; i < GRIDSIZE_Y; i++)
		{
			tempf += d_prob_y[idx_y*GRIDSIZE_Y + i] * d_VNDupdate[i*GRIDSIZE_B + idx_b];
		}
		d_ceC[idx_y*GRIDSIZE_B + idx_b] = tempf;

		// def prob
		tempf = 0;
		for (i = 0; i < GRIDSIZE_Y; i++)
		{
			tempf += d_prob_y[idx_y*GRIDSIZE_Y+i]* d_defprob_update[i*GRIDSIZE_B + idx_b];
		}
		d_Edefprob[idx_y*GRIDSIZE_B + idx_b] = tempf;

		// qND_H
		tempf = 0;
		for (i = 0; i < GRIDSIZE_Y; i++)
		{
			tempf += d_prob_y[idx_y*GRIDSIZE_Y+i]* d_qH_NDupdate[i*GRIDSIZE_B + idx_b];
		}
		d_qH_ND[idx_y*GRIDSIZE_B + idx_b] = tempf / (1 + RH);

		// qND_L
		tempf = 0;
		for (i = 0; i < GRIDSIZE_Y; i++)
		{
			tempf += d_prob_y[idx_y*GRIDSIZE_Y + i] * d_qL_NDupdate[i*GRIDSIZE_B + idx_b];
		}
		d_qL_ND[idx_y*GRIDSIZE_B + idx_b] = tempf / (1 + RL);

		// qD_H
		tempf = 0;
		for (i = 0; i < GRIDSIZE_Y; i++)
		{
			tempf += d_prob_y[idx_y*GRIDSIZE_Y + i] * ((1-PROB_HL)*d_qH_D0[i*GRIDSIZE_B + idx_b] + PROB_HL*d_qL_D0[i*GRIDSIZE_B + idx_b]);
		}
		d_qH_D1[idx_y*GRIDSIZE_B + idx_b] = tempf;

		// qD_L
		tempf = 0;
		for (i = 0; i < GRIDSIZE_Y; i++)
		{
			tempf += d_prob_y[idx_y*GRIDSIZE_Y + i] * max(0.0, (1 - PROB_LH_D)*d_qL_D0[i*GRIDSIZE_B + idx_b] + PROB_LH_D*d_qH_D0[i*GRIDSIZE_B + idx_b]  - HOLDING_COST);
		}
		d_qL_D1[idx_y*GRIDSIZE_B + idx_b] = tempf;

	}
}

__global__ void vfi_update2(REAL_TYPE *d_CVD, REAL_TYPE *d_CVDhaircut, REAL_TYPE *d_CVNDhaircut,
	REAL_TYPE *d_qH_D, REAL_TYPE *d_qL_D, REAL_TYPE *d_qH_Dhaircut, REAL_TYPE *d_qL_Dhaircut,
	REAL_TYPE *d_qH_NDhaircut, REAL_TYPE *d_qL_NDhaircut)
// update taking into account write downs
{
	int i = blockIdx.x*blockDim.x + threadIdx.x;
	if (i < GRIDSIZE_Y*GRIDSIZE_B)
	{
		int idx_b = i % GRIDSIZE_B;
		d_CVD[i] = (1 - REENTERPROB)*d_CVDhaircut[i] + REENTERPROB*d_CVNDhaircut[i];
		d_qH_D[i] = max(0.0,(1 - REENTERPROB)*d_qH_Dhaircut[i] / (1 + RH)
			+ GRID_RECOVERY_FRAC[idx_b] * REENTERPROB * d_qH_NDhaircut[i]);
		d_qL_D[i] = max(0.0,(1 - REENTERPROB)*d_qL_Dhaircut[i] / (1 + RL)
			+ GRID_RECOVERY_FRAC[idx_b] * REENTERPROB * d_qL_NDhaircut[i]);
	}
}

__global__ void update_compute_errors(REAL_TYPE *d_x, REAL_TYPE *d_y, REAL_TYPE wx)
// new variable stored in x
// errors stored in y
{
	int i = blockIdx.x*blockDim.x + threadIdx.x;
	if (i < GRIDSIZE_Y*GRIDSIZE_B)
	{
		REAL_TYPE x = d_x[i];
		REAL_TYPE y = d_y[i];
		d_x[i] = wx*x + (1 - wx)*y;
		// d_y[i] = fabsf(x - y);
		int idx_b = i % GRIDSIZE_B;
		int idx_y = (i - idx_b)/GRIDSIZE_B;
		if (idx_y>=CONV_CHK_TFP_LB_IDX && idx_y<=CONV_CHK_TFP_UB_IDX && idx_b<= CONV_CHK_TFP_UB_IDX)
		{
			d_y[i] = abs(x - y);
		}
		else
		{
			d_y[i] = 0.0;
		}
	}
}

void vfi(parms_bsl_mod &p, REAL_TYPE wOldV, REAL_TYPE wOldQ, REAL_TYPE wOldDefProb, REAL_TYPE *d_prob_y, REAL_TYPE *d_uD_yb, 
	REAL_TYPE &err_CVD, REAL_TYPE &err_CVND, REAL_TYPE &err_qH_ND, REAL_TYPE &err_qL_ND, REAL_TYPE &err_qH_D, REAL_TYPE &err_qL_D, REAL_TYPE &err_defprob,
	REAL_TYPE *d_CVD, REAL_TYPE *d_CVND, REAL_TYPE *d_qH_ND, REAL_TYPE *d_qL_ND, REAL_TYPE *d_qH_D, REAL_TYPE *d_qL_D, REAL_TYPE *d_defprob, REAL_TYPE *d_VD, 
	REAL_TYPE *d_VNDupdate, REAL_TYPE *d_qH_ND_update, REAL_TYPE *d_qL_ND_update, REAL_TYPE *d_defprob_update, REAL_TYPE *d_defthresh,
	REAL_TYPE *d_EVD, REAL_TYPE *d_CVNDnew,
	REAL_TYPE *d_qH_NDnew, REAL_TYPE *d_qL_NDnew, REAL_TYPE *d_EqH_D, REAL_TYPE *d_EqL_D, REAL_TYPE *d_defprob_new,
	REAL_TYPE *d_CVNDi, REAL_TYPE *d_qH_NDi, REAL_TYPE *d_qL_NDi,
	REAL_TYPE *d_CVDnew, REAL_TYPE *d_qH_Dnew, REAL_TYPE *d_qL_Dnew)
	/*
	All vectors should have dimension GRIDSIZE_Y*GRIDSIZE_B
	// the device has plenty of memory
	*/
{

	int NUMTHREADS1 = 128;
	int NUMBLOCKS1 = (int) ceil((double)(p.gridsize_tfp*p.gridsize_b) / (double)NUMTHREADS1);

	// temp1 = VD, temp2 = VNDupdate, temp3 = qHNDupdate, temp4 = qLNDupdate
	vfi_iterate <<< NUMBLOCKS1, NUMTHREADS1 >>> (d_VD, d_VNDupdate, d_qH_ND_update, d_qL_ND_update, d_defprob_update, d_defthresh,
		d_CVD, d_qH_ND, d_qL_ND,d_qH_D, d_qL_D, d_CVND, d_uD_yb, d_defprob); // inputs
	cudaDeviceSynchronize();

	// first round of updating
	int NUMTHREADS2 = 128;
	int NUMBLOCKS2 = (int)ceil((double)(p.gridsize_tfp*p.gridsize_b) / (double)NUMTHREADS2);

	vfi_update1 <<< NUMBLOCKS2, NUMTHREADS2 >>> (d_prob_y, d_EVD, d_CVNDnew, d_qH_NDnew, d_qL_NDnew, d_EqH_D, d_EqL_D, d_defprob_new,
		d_VD, d_VNDupdate, d_qH_ND_update, d_qL_ND_update, d_defprob_update, d_qH_D, d_qL_D);
	cudaDeviceSynchronize();

	// interpolate

	// temp1 = ceDi, temp2 = qH_Di, temp3 = qL_Di, temp4 = ceCi, temp11 = qH_NDi, temp12 = qL_NDi
	vfi_interpolate <<< NUMBLOCKS2, NUMTHREADS2 >>> (d_CVNDi, d_qH_NDi, d_qL_NDi,
		d_CVNDnew, d_qH_NDnew, d_qL_NDnew);
	cudaDeviceSynchronize();

	// second round of updating

	// temp5 = CVD***, temp9 = qH_D***, temp10 = qL_D***
	vfi_update2 <<< NUMBLOCKS2, NUMTHREADS2 >>> (d_CVDnew, d_EVD, d_CVNDi,
		d_qH_Dnew, d_qL_Dnew, d_EqH_D, d_EqL_D,
		d_qH_NDi, d_qL_NDi);
	cudaDeviceSynchronize();

	// update and compute errors
	int NUMTHREADS3 = 128;
	int NUMBLOCKS3 = (int)ceil((double)(p.gridsize_tfp*p.gridsize_b) / (double)NUMTHREADS3);

	update_compute_errors <<< NUMBLOCKS3, NUMTHREADS3 >>> (d_CVD, d_CVDnew, wOldV); // CVD
	update_compute_errors <<< NUMBLOCKS3, NUMTHREADS3 >>> (d_CVND, d_CVNDnew, wOldV); // CVND
	update_compute_errors <<< NUMBLOCKS3, NUMTHREADS3 >>> (d_defprob, d_defprob_new, wOldDefProb); // def prob
	update_compute_errors <<< NUMBLOCKS3, NUMTHREADS3 >>> (d_qH_ND, d_qH_NDnew, wOldQ); // qH_ND
	update_compute_errors <<< NUMBLOCKS3, NUMTHREADS3 >>> (d_qL_ND, d_qL_NDnew, wOldQ); // qL_ND
	update_compute_errors <<< NUMBLOCKS3, NUMTHREADS3 >>> (d_qH_D, d_qH_Dnew, wOldQ); // qH_D
	update_compute_errors <<< NUMBLOCKS3, NUMTHREADS3 >>> (d_qL_D, d_qL_Dnew, wOldQ); // qL_D

	cudaDeviceSynchronize();

	// reduce for norms
	thrust::device_ptr<REAL_TYPE> tempptr1 = thrust::device_pointer_cast(d_CVDnew);
	thrust::device_vector<REAL_TYPE>::iterator tempiter1 = thrust::max_element(tempptr1, tempptr1 + p.gridsize_tfp*p.gridsize_b);
	err_CVD = *tempiter1;

	thrust::device_ptr<REAL_TYPE> tempptr2 = thrust::device_pointer_cast(d_CVNDnew);
	thrust::device_vector<REAL_TYPE>::iterator tempiter2 = thrust::max_element(tempptr2, tempptr2 + p.gridsize_tfp*p.gridsize_b);
	err_CVND = *tempiter2;

	thrust::device_ptr<REAL_TYPE> tempptr3 = thrust::device_pointer_cast(d_qH_NDnew);
	thrust::device_vector<REAL_TYPE>::iterator tempiter3 = thrust::max_element(tempptr3, tempptr3 + p.gridsize_tfp*p.gridsize_b);
	err_qH_ND = *tempiter3;

	// int idx_maxerr_qH_ND = &tempiter3[0] - &tempptr3[0];
	// int idx_maxerr_qH_ND_b = idx_maxerr_qH_ND % p.gridsize_b;
	// int idx_maxerr_qH_ND_y = (idx_maxerr_qH_ND - idx_maxerr_qH_ND_b)/p.gridsize_b;
	// std::cout<< "maxerr(qH_ND) occurred at idx_y=" << idx_maxerr_qH_ND_y << ", idx_b=" << idx_maxerr_qH_ND_b << std::endl;

	thrust::device_ptr<REAL_TYPE> tempptr4 = thrust::device_pointer_cast(d_qL_NDnew);
	thrust::device_vector<REAL_TYPE>::iterator tempiter4 = thrust::max_element(tempptr4, tempptr4 + p.gridsize_tfp*p.gridsize_b);
	err_qL_ND = *tempiter4;

	thrust::device_ptr<REAL_TYPE> tempptr5 = thrust::device_pointer_cast(d_qH_Dnew);
	thrust::device_vector<REAL_TYPE>::iterator tempiter5 = thrust::max_element(tempptr5, tempptr5 + p.gridsize_tfp*p.gridsize_b);
	err_qH_D = *tempiter5;

	thrust::device_ptr<REAL_TYPE> tempptr6 = thrust::device_pointer_cast(d_qL_Dnew);
	thrust::device_vector<REAL_TYPE>::iterator tempiter6 = thrust::max_element(tempptr6, tempptr6 + p.gridsize_tfp*p.gridsize_b);
	err_qL_D = *tempiter6;

	thrust::device_ptr<REAL_TYPE> tempptr7 = thrust::device_pointer_cast(d_defprob_new);
	thrust::device_vector<REAL_TYPE>::iterator tempiter7 = thrust::max_element(tempptr7, tempptr7 + p.gridsize_tfp*p.gridsize_b);
	err_defprob = *tempiter7;

	cudaDeviceSynchronize();

}

__global__ void initValueFuns(REAL_TYPE *d_CVD, REAL_TYPE *d_CVND, REAL_TYPE *d_qH_ND, REAL_TYPE *d_qL_ND, REAL_TYPE *d_qH_D, REAL_TYPE *d_qL_D,
	REAL_TYPE *d_defprob)
{
	int i = blockIdx.x*blockDim.x + threadIdx.x;
	if (i < GRIDSIZE_Y*GRIDSIZE_B)
	{
		d_CVD[i] = 0;
		d_CVND[i] = 0;
		d_qH_ND[i] = 0;
		d_qL_ND[i] = 0;
		d_qH_D[i] = 0;
		d_qL_D[i] = 0;
		d_defprob[i] = 0;
	}
}

int SolveModel(parms_bsl_mod &p, const char* save_folder_name, int useDevice)
{

	p.displayparms();

	// initialize gpu
	int numDevices = 0;
	cudaGetDeviceCount(&numDevices);
	if (useDevice >= numDevices)
	{
		printf("Error: program requested device %d; only %d device(s) are available.\n", useDevice, numDevices);
		exit(EXIT_FAILURE);
	}
	cudaSetDevice(useDevice);

	p.initDeviceConstants(); // this calls rouwenhorst.

	if (p.checkparms() == EXIT_FAILURE)
	{
		std::cout << "Error with parameter(s)." << std::endl;
		exit(EXIT_FAILURE);
	}

	// debug
	printDeviceConstants <<< 1,1 >>> ();
	cudaDeviceSynchronize();

	// allocate device memory
	REAL_TYPE *d_CVD, *d_CVND, *d_qH_ND, *d_qL_ND, *d_qH_D, *d_qL_D, *d_defprob;
	REAL_TYPE *d_VD, *d_VNDupdate, *d_qH_ND_update, *d_qL_ND_update, *d_EVD, *d_CVNDnew,
		*d_qH_NDnew, *d_qL_NDnew, *d_EqH_D, *d_EqL_D,
		*d_CVNDi, *d_qH_NDi, *d_qL_NDi,
		*d_CVDnew, *d_qH_Dnew, *d_qL_Dnew, *d_defprob_new, *d_defprob_update, *d_defthresh;
	REAL_TYPE *d_prob_y, *d_uD_yb;

	cudaMalloc((void**)&d_defprob, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_defprob_new, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_defprob_update, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_defthresh, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_CVD, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_CVND, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_qH_ND, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_qL_ND, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_qH_D, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_qL_D, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_VD, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_VNDupdate, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_qH_ND_update, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_qL_ND_update, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_EVD, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_CVNDnew, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_qH_NDnew, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_qL_NDnew, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_EqH_D, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_EqL_D, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_CVNDi, p.gridsize_tfp*p.gridsize_b * sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_qH_NDi, p.gridsize_tfp*p.gridsize_b * sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_qL_NDi, p.gridsize_tfp*p.gridsize_b * sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_CVDnew, p.gridsize_tfp*p.gridsize_b * sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_qH_Dnew, p.gridsize_tfp*p.gridsize_b * sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_qL_Dnew, p.gridsize_tfp*p.gridsize_b * sizeof(REAL_TYPE));
	cudaMalloc((void**)&d_prob_y, p.gridsize_tfp*p.gridsize_tfp*sizeof(REAL_TYPE));
	cudaMemcpy(d_prob_y, p.prob_y, p.gridsize_tfp*p.gridsize_tfp*sizeof(REAL_TYPE), cudaMemcpyHostToDevice);
	cudaMalloc((void**)&d_uD_yb, p.gridsize_tfp*p.gridsize_b * sizeof(REAL_TYPE));
	cudaMemcpy(d_uD_yb, p.grid_uD_yb, p.gridsize_tfp*p.gridsize_b * sizeof(REAL_TYPE), cudaMemcpyHostToDevice);

	initValueFuns <<< (int)ceil((double)(p.gridsize_tfp*p.gridsize_b)/(double)128) , 128 >>> (d_CVD, d_CVND, d_qH_ND, d_qL_ND, d_qH_D, 
		d_qL_D, d_defprob);
	cudaDeviceSynchronize();

	size_t free_cuda_mem, total_cuda_mem;
	cudaMemGetInfo(&free_cuda_mem, &total_cuda_mem);
	std::cout << "Memory allocated for value function iteration. Total GPU memory: " << round(100 * (double)total_cuda_mem / 1073741824) / 100
		<< "GBs. Free memory: " << round(100 * (double)free_cuda_mem / 1073741824) / 100 << "GBs." << std::endl;

	cudaEvent_t start, stop;
	float milliseconds_taken;
	cudaEventCreate(&start);
	cudaEventCreate(&stop);
	cudaEventRecord(start);

	int iter = 0;
	int Updatek = 0;
	bool bConverged = false;
	int DisplayInterval = 25;
	REAL_TYPE err_CVD, err_CVND, err_qH_ND, err_qL_ND, err_qH_D, err_qL_D, err_q, err_defprob;
	REAL_TYPE wOldV, wOldQ, wOldDefProb;

	do
	{
		if (Updatek < p.NumUpdateRegimes - 1)
		{
			if (iter >= p.UpdateRegimeIter[Updatek])
			{
				Updatek += 1;
			}
		}

		iter += 1;
		wOldV = p.UpdateWeightOldV[Updatek];
		wOldQ = p.UpdateWeightOldQ[Updatek];
		wOldDefProb = p.UpdateWeightOldDefProb[Updatek];
		vfi(p, wOldV, wOldQ, wOldDefProb, d_prob_y, d_uD_yb, 
			err_CVD, err_CVND, err_qH_ND, err_qL_ND, err_qH_D, err_qL_D, err_defprob,
			d_CVD, d_CVND, d_qH_ND, d_qL_ND, d_qH_D, d_qL_D, d_defprob,
			d_VD, d_VNDupdate, d_qH_ND_update, d_qL_ND_update, d_defprob_update, d_defthresh,
			d_EVD, d_CVNDnew,
			d_qH_NDnew, d_qL_NDnew, d_EqH_D, d_EqL_D, d_defprob_new,
			d_CVNDi, d_qH_NDi, d_qL_NDi,
			d_CVDnew, d_qH_Dnew, d_qL_Dnew);

		err_q = err_qH_ND;
		if (err_qL_ND > err_q)
		{
			err_q = err_qL_ND;
		}
		if (err_qH_D > err_q)
		{
			err_q = err_qH_D;
		}
		if (err_qL_D > err_q)
		{
			err_q = err_qL_D;
		}
		if (err_defprob > err_q)
		{
			err_q = err_defprob;
		}

		bConverged = err_q <= p.errtolQ && err_CVD <= p.errtolV && err_CVND <= p.errtolV;

		if (iter % DisplayInterval == 0)
		{
			cudaEventRecord(stop);
			cudaEventSynchronize(stop);
			cudaEventElapsedTime(&milliseconds_taken, start, stop);
			std::cout << "iter " << iter << ", err(defprob)=" << err_defprob << ", err(qs)=" << err_q << ", err(CVD)=" << err_CVD << ", err(CVND)=" << err_CVND << ". " << milliseconds_taken / 1000.0 << " seconds elapsed." << std::endl;
		}

	} while (iter <= p.maxiters && !bConverged);

	// first save
	cudaDeviceSynchronize();
	REAL_TYPE *h_CVD_old, *h_CVND_old, *h_qH_ND_old, *h_qL_ND_old, *h_qH_D_old, *h_qL_D_old, *h_defprob_old;
	h_defprob_old = new REAL_TYPE[p.gridsize_tfp*p.gridsize_b];
	h_CVD_old = new REAL_TYPE[p.gridsize_tfp*p.gridsize_b];
	h_CVND_old = new REAL_TYPE[p.gridsize_tfp*p.gridsize_b];
	h_qH_ND_old = new REAL_TYPE[p.gridsize_tfp*p.gridsize_b];
	h_qL_ND_old = new REAL_TYPE[p.gridsize_tfp*p.gridsize_b];
	h_qH_D_old = new REAL_TYPE[p.gridsize_tfp*p.gridsize_b];
	h_qL_D_old = new REAL_TYPE[p.gridsize_tfp*p.gridsize_b];
	cudaMemcpy(h_defprob_old,d_defprob, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE), cudaMemcpyDeviceToHost);
	cudaMemcpy(h_CVD_old,d_CVD, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE), cudaMemcpyDeviceToHost);
	cudaMemcpy(h_CVND_old, d_CVND, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE), cudaMemcpyDeviceToHost);
	cudaMemcpy(h_qH_ND_old, d_qH_ND, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE), cudaMemcpyDeviceToHost);
	cudaMemcpy(h_qL_ND_old, d_qL_ND, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE), cudaMemcpyDeviceToHost);
	cudaMemcpy(h_qH_D_old, d_qH_D, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE), cudaMemcpyDeviceToHost);
	cudaMemcpy(h_qL_D_old, d_qL_D, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE), cudaMemcpyDeviceToHost);

	// do a last run
	// do a quick hack
	cudaDeviceSynchronize();
	wOldV = 0.0;
	wOldQ = 0.0;
	wOldDefProb = 0.0;
	vfi(p, wOldV, wOldQ, wOldDefProb, d_prob_y, d_uD_yb, 
		err_CVD, err_CVND, err_qH_ND, err_qL_ND, err_qH_D, err_qL_D, err_defprob,
		d_CVD, d_CVND, d_qH_ND, d_qL_ND, d_qH_D, d_qL_D, d_defprob,
		d_VD, d_VNDupdate, d_qH_ND_update, d_qL_ND_update, d_defprob_update, d_defthresh,
		d_EVD, d_CVNDnew,
		d_qH_NDnew, d_qL_NDnew, d_EqH_D, d_EqL_D, d_defprob_new,
		d_CVNDi, d_qH_NDi, d_qL_NDi,
		d_CVDnew, d_qH_Dnew, d_qL_Dnew);
	cudaDeviceSynchronize();

	cudaEventRecord(stop);
	cudaEventSynchronize(stop);
	cudaEventElapsedTime(&milliseconds_taken, start, stop);
	std::cout << "Program stopped after " << iter << " iterations. " << milliseconds_taken / 1000.0 << " seconds elapsed." << std::endl;
	std::cout << "iter " << iter << ", err(defprob)=" << err_defprob << ", err(qs)=" << err_q << ", err(CVD)=" << err_CVD << ", err(CVND)=" << err_CVND << ". " << milliseconds_taken / 1000.0 << " seconds elapsed." << std::endl;

	// save outputs
	REAL_TYPE *h_CVD, *h_CVND, *h_qH_ND, *h_qL_ND, *h_qH_D, *h_qL_D, *h_defprob, *h_defthresh;
	h_defprob = new REAL_TYPE[p.gridsize_tfp*p.gridsize_b];
	h_defthresh = new REAL_TYPE[p.gridsize_tfp*p.gridsize_b];
	h_CVD = new REAL_TYPE[p.gridsize_tfp*p.gridsize_b];
	h_CVND = new REAL_TYPE[p.gridsize_tfp*p.gridsize_b];
	h_qH_ND = new REAL_TYPE[p.gridsize_tfp*p.gridsize_b];
	h_qL_ND = new REAL_TYPE[p.gridsize_tfp*p.gridsize_b];
	h_qH_D = new REAL_TYPE[p.gridsize_tfp*p.gridsize_b];
	h_qL_D = new REAL_TYPE[p.gridsize_tfp*p.gridsize_b];
	cudaMemcpy(h_defprob,d_defprob, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE), cudaMemcpyDeviceToHost);
	cudaMemcpy(h_CVD,d_CVD, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE), cudaMemcpyDeviceToHost);
	cudaMemcpy(h_CVND, d_CVND, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE), cudaMemcpyDeviceToHost);
	cudaMemcpy(h_qH_ND, d_qH_ND, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE), cudaMemcpyDeviceToHost);
	cudaMemcpy(h_qL_ND, d_qL_ND, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE), cudaMemcpyDeviceToHost);
	cudaMemcpy(h_qH_D, d_qH_D, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE), cudaMemcpyDeviceToHost);
	cudaMemcpy(h_qL_D, d_qL_D, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE), cudaMemcpyDeviceToHost);
	cudaMemcpy(h_defthresh, d_defthresh, p.gridsize_tfp*p.gridsize_b*sizeof(REAL_TYPE), cudaMemcpyDeviceToHost);

	// std::string save_folder_name2 = save_outputs(save_folder_name, p, h_CVD, h_CVND, h_qH_ND, h_qL_ND, h_qH_D, h_qL_D, 
	// 	iter, err_CVD, err_CVND, err_qH_ND, err_qL_ND, err_qH_D, err_qL_D);

	std::string save_folder_name2 = save_outputs_detailed(save_folder_name, p, h_CVD, h_CVND, h_qH_ND, h_qL_ND, h_qH_D, h_qL_D, 
		iter, err_CVD, err_CVND, err_qH_ND, err_qL_ND, err_qH_D, err_qL_D,
		h_CVD_old, h_CVND_old, h_qH_ND_old,
		h_qL_ND_old, h_qH_D_old, h_qL_D_old, h_defprob, h_defprob_old, h_defthresh);

	std::cout << "results saved to folder " << save_folder_name2.c_str() << std::endl;

	delete[] h_defprob;
	delete[] h_defthresh;
	delete[] h_CVD;
	delete[] h_CVND;
	delete[] h_qH_ND;
	delete[] h_qL_ND;
	delete[] h_qH_D;
	delete[] h_qL_D;

	delete[] h_defprob_old;
	delete[] h_CVD_old;
	delete[] h_CVND_old;
	delete[] h_qH_ND_old;
	delete[] h_qL_ND_old;
	delete[] h_qH_D_old;
	delete[] h_qL_D_old;

	// free device memory
	cudaFree(d_defprob);
	cudaFree(d_defprob_update);
	cudaFree(d_defthresh);
	cudaFree(d_defprob_new);
	cudaFree(d_CVD);
	cudaFree(d_CVND);
	cudaFree(d_qH_ND);
	cudaFree(d_qL_ND);
	cudaFree(d_qH_D);
	cudaFree(d_qL_D);
	cudaFree(d_VD);
	cudaFree(d_VNDupdate);
	cudaFree(d_qH_ND_update);
	cudaFree(d_qL_ND_update);
	cudaFree(d_EVD);
	cudaFree(d_CVNDnew);
	cudaFree(d_qH_NDnew);
	cudaFree(d_qL_NDnew);
	cudaFree(d_EqH_D);
	cudaFree(d_EqL_D);
	cudaFree(d_CVNDi);
	cudaFree(d_qH_NDi);
	cudaFree(d_qL_NDi);
	cudaFree(d_CVDnew);
	cudaFree(d_qH_Dnew);
	cudaFree(d_qL_Dnew);
	cudaFree(d_prob_y);

	return EXIT_SUCCESS;

}